# Copyright 2019-2023 Quantinuum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
The taskgraph module defines the TaskGraph, MitRes and MitTask objects, through which
mitigated experiments are run.
"""

from .measurement_reduction import gen_MeasurementReduction_MitEx
from .mitex import (
    MitEx,
    collate_circuit_shots_task_gen,
    filter_observable_tracker_task_gen,
    gen_compiled_MitRes,
    gen_compiled_shot_split_MitRes,
    get_expectations_task_gen,
    split_results_task_gen,
)
from .mitres import (
    MitRes,
    backend_compile_circuit_shots_task_gen,
    backend_handle_task_gen,
    backend_res_task_gen,
    gen_shot_split_MitRes,
    group_shots_task_gen,
    split_shots_task_gen,
)
from .mittask import (
    AnsatzCircuit,
    CircuitShots,
    IOTask,
    MitTask,
    ObservableExperiment,
    Wire,
    duplicate_wire_task_gen,
)
from .task_graph import TaskGraph
from .utils import MeasurementCircuit, ObservableTracker, SymbolsDict



# Copyright 2019-2023 Quantinuum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
qermit provides tools for running and composing error-mitigation methods. Error
mitigation methods are split into two types, those which in some manner modify the set
of shots returned when running quantum circuits on quantum devices (MitRes), and those which
modify the expectation value of some observable (MitEx).
"""

from qermit.taskgraph.mitex import MitEx
from qermit.taskgraph.mitres import MitRes
from qermit.taskgraph.mittask import (
    AnsatzCircuit,
    CircuitShots,
    MitTask,
    ObservableExperiment,
)
from qermit.taskgraph.task_graph import TaskGraph
from qermit.taskgraph.utils import MeasurementCircuit, ObservableTracker, SymbolsDict

__path__ = __import__("pkgutil").extend_path(__path__, __name__)



# Copyright 2019-2023 Quantinuum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
The noise_model module provides an assortment of noise modeling and simulation
tools for local testing.
"""

from .mock_quantinuum_backend import MockQuantinuumBackend
from .noise_model import (
    Direction,
    ErrorDistribution,
    LogicalErrorDistribution,
    NoiseModel,
)
from .pauli_error_transpile import PauliErrorTranspile
from .qermit_pauli import QermitPauli
from .transpiler_backend import TranspilerBackend



# Copyright 2019-2023 Quantinuum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import operator
from copy import copy
from enum import Enum
from functools import reduce
from math import ceil
from typing import Counter, List, Sequence, Tuple, cast

from pytket import Bit, Circuit, OpType
from pytket.backends import Backend
from pytket.backends.backendresult import BackendResult
from pytket.passes import AutoRebase
from pytket.tailoring import PauliFrameRandomisation, UniversalFrameRandomisation
from pytket.utils.outcomearray import OutcomeArray

from qermit import (
    CircuitShots,
    MitRes,
    MitTask,
)
from qermit.taskgraph.mitex import backend_compile_circuit_shots_task_gen


class FrameRandomisation(Enum):
    @staticmethod
    def PauliFrameRandomisation(
        circuit: Circuit, shots: int, samples: int
    ) -> List[CircuitShots]:
        """
        Uses the pytket.tailoring method PauliFrameRandomisation  to return a list of
        circuits corresponding to instances of frame randomisation for the input circuit.

        :param circuit: Circuit to have frame randomisation applied.
        :param shots: Number of shots of each frame randomisation circuit to take.
        :param samples: Number of frame randomisation instances to sample.
        """
        pfr = PauliFrameRandomisation()
        pfr_shots = ceil(shots / samples)
        AutoRebase({OpType.CX, OpType.Rz, OpType.H, OpType.S}).apply(circuit)
        pfr_circuits = pfr.sample_circuits(circuit, samples)
        return [CircuitShots(Circuit=c, Shots=pfr_shots) for c in pfr_circuits]

    @staticmethod
    def UniversalFrameRandomisation(
        circuit: Circuit, shots: int, samples: int
    ) -> List[CircuitShots]:
        """
        Uses the pytket.tailoring method UniversalFrameRandomisation  to return a list of
        circuits corresponding to instances of frame randomisation for the input circuit.
        UniversalFrameRandomisation rebases the circuit to a gate set such that
        the whole circuit is a single cycle for frame randomisation. This is possible with
        the additional noise assumption on top of the regular Frame Randomisation assumptions
        that Rz(-x) and Rz(x) incur similar noise for any angle x.

        :param circuit: Circuit to have frame randomisation applied.
        :param shots: Number of shots of each frame randomisation circuit to take.
        :param samples: Number of frame randomisation instances to sample.
        """
        ufr = UniversalFrameRandomisation()
        ufr_shots = ceil(shots / samples)

        AutoRebase({OpType.CX, OpType.Rz, OpType.H}).apply(circuit)
        ufr_circuits = ufr.sample_circuits(circuit, samples)
        return [CircuitShots(Circuit=c, Shots=ufr_shots) for c in ufr_circuits]


def frame_randomisation_circuits_task_gen(
    samples: int, _fr_type: FrameRandomisation
) -> MitTask:
    """
    Returns a MitTask object that produces Frame Randomisation circuits for some wire of experiment circuits.

    :param samples: Number of samples of frame randomisation circuits to take for each circuit in wire.
    """

    def task(obj, circs_shots: List[CircuitShots]) -> Tuple[List[CircuitShots]]:
        """
        :param circ_shots: A list of tuple of circuits and shots. Each circuit has frame randomisation applied.
        The number of shots of each frame randomisation circuit is ceil(shots/samples)

        :return: Frame Randomisation circutis
        """

        all_fr_circs_shots = []
        for circ, shots in circs_shots:
            fr_circshots = _fr_type(circ, shots, samples)  # type: ignore
            # frame randomisation type not callable as no default enum
            all_fr_circs_shots.extend(fr_circshots)
        return (all_fr_circs_shots,)

    return MitTask(
        _label="GenerateFrameRandomisationCircuits",
        _n_in_wires=1,
        _n_out_wires=1,
        _method=task,
    )


def frame_randomisation_result_task_gen(samples: int) -> MitTask:
    """
    Returns a MitTask object that sequentially collates samples number of BackendResult objects into single
    BackendResult object. These collated BackendResult objects should include all frame experiments for a single
    original circuit.

    :param samples: Number of frame randomisation instances created in the first place, used to suitably
        collate results.
    """

    def task(
        obj,
        all_fr_results: List[BackendResult],
    ) -> Tuple[List[BackendResult]]:
        """
        :param all_fr_results: A list of BackendResult objects for all frame randomisations for all experiment circuits.

        :return: Collated BackendResult objects, reflecting frame randomisation procedure
        """
        chunked_results = [
            all_fr_results[i : i + samples]
            for i in range(0, len(all_fr_results), samples)
        ]
        results = []
        for chunk in chunked_results:
            all_counts = [result.get_counts() for result in chunk]
            combined_counts = reduce(operator.add, all_counts)
            outcome_array = {
                OutcomeArray.from_readouts([key]): val
                for key, val in combined_counts.items()
            }
            outcome_counts = Counter(outcome_array)
            results.append(
                BackendResult(
                    counts=outcome_counts, c_bits=cast(Sequence[Bit], chunk[0].c_bits)
                )
            )
        return (results,)

    return MitTask(
        _label="CollateFrameRandomisationResults",
        _n_in_wires=1,
        _n_out_wires=1,
        _method=task,
    )


def gen_Frame_Randomisation_MitRes(backend: Backend, samples: int, **kwargs) -> MitRes:
    """
    Produces a MitRes object that applies FrameRandomisation techniques to experiment circuits.

    :param backend: Backend which experiments are default run through.
    :param samples: Number of Frame Randomisation instances sampled in Frame Randomisation.

    :key mitres: MitRes object FrameRandomisation MitRes built around if given.
    :key frame_randomisation: FrameRandomisation Enum passed to specify method used.
        Default set to FrameRandomisation.UniversalFrameRandomisation.
    """
    _mitres = copy(
        kwargs.get("mitres", MitRes(backend, _label="FrameRandomisationMitRes"))
    )
    _fr_type = copy(
        kwargs.get(
            "frame_randomisation", FrameRandomisation.UniversalFrameRandomisation
        )
    )
    _mitres.prepend(backend_compile_circuit_shots_task_gen(backend))
    _mitres.prepend(frame_randomisation_circuits_task_gen(samples, _fr_type))
    _mitres.append(
        frame_randomisation_result_task_gen(samples),
    )
    for n in _mitres._task_graph.nodes:
        if hasattr(n, "_label"):
            n._label = "FR" + n._label
    return _mitres



# Copyright 2019-2023 Quantinuum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import itertools
from collections import OrderedDict, namedtuple
from enum import Enum
from functools import lru_cache
from math import ceil, log2
from typing import Counter, Dict, Iterable, List, Optional, Tuple, cast

import numpy as np
from pytket.backends import Backend
from pytket.backends.backendresult import BackendResult
from pytket.circuit import Bit, CircBox, Circuit, Node, Qubit
from pytket.passes import DecomposeBoxes, FlattenRegisters
from pytket.unit_id import UnitID
from pytket.utils.outcomearray import OutcomeArray

FullCorrelatedNoiseCharacterisation = namedtuple(
    "FullCorrelatedNoiseCharacterisation",
    ["CorrelatedNodes", "NodeToIntDict", "CharacterisationMatrices"],
)

StateInfo = namedtuple("StateInfo", ["PreparedStates", "QubitBitMaps"])


# Helper methods for holding basis states
@lru_cache(maxsize=128)
def binary_to_int(bintuple: Tuple[int]) -> int:
    """Convert a binary tuple to corresponding integer, with most significant bit as the
    first element of tuple.

    :param bintuple: Binary tuple
    :return: Integer
    """
    integer = 0
    for index, bitset in enumerate(reversed(bintuple)):
        if bitset:
            integer |= 1 << index
    return integer


@lru_cache(maxsize=128)
def int_to_binary(val: int, dim: int) -> Tuple[int, ...]:
    """Convert an integer to corresponding binary tuple, with most significant bit as
    the first element of tuple.

    :param val: input integer
    :param dim: Bit width
    :return: Binary tuple of width dim
    """
    return tuple(map(int, format(val, "0{}b".format(dim))))


def get_full_transition_tomography_circuits(
    process_circuit: Circuit, backend: Backend, correlations: List[List[Node]]
) -> Tuple[List[Circuit], List[StateInfo]]:
    """Generate calibration circuits according to the specified correlation, backend and given circuit.

    :param circuit: Circuit surmising correlated noise process being characterised
    :param backend: Backend on which the experiments are run.
    :param correlations: A list of lists of correlated Nodes of a `Device`.
        Qubits within the same list are assumed to only have errors correlated
        with each other. Thus to allow errors between all qubits you should
        provide a single list.  The qubits in `correlations` must be nodes in the
        backend's associated `Device`.
    :return: A list of calibration circuits to be run on the machine. The circuits
        should be processed without compilation.
    """

    def to_tuple(correlation_list: List[Node]) -> Tuple[Node, ...]:
        return tuple(correlation_list)

    subsets_matrix_map = OrderedDict.fromkeys(
        sorted(map(to_tuple, correlations), key=len, reverse=True)
    )
    # ordered from largest to smallest via OrderedDict & sorted
    subset_dimensions = [len(subset) for subset in subsets_matrix_map]
    major_state_dimensions = subset_dimensions[0]
    n_circuits = 1 << major_state_dimensions
    all_qubits = [qb for subset in correlations for qb in subset]

    if len(process_circuit.qubits) != len(all_qubits):
        raise ValueError(
            "Process being characterised has {} qubits, correlations only specify {} qubits.".format(
                len(process_circuit.qubits), len(all_qubits)
            )
        )

    # output
    prepared_circuits = []
    state_infos = []

    # set up CircBox of X gate for preparing basis states
    xcirc = Circuit(1).X(0)
    xcirc = backend.get_compiled_circuit(xcirc, optimisation_level=0)
    FlattenRegisters().apply(xcirc)
    xbox = CircBox(xcirc)
    # need to be default register to add as box suitably

    n_qubits_pre_compile = process_circuit.n_qubits
    # This needs to be optimisation level 0 to avoid using simplify initial
    process_circuit = backend.get_compiled_circuit(
        process_circuit, optimisation_level=0
    )

    while process_circuit.n_qubits < n_qubits_pre_compile:
        process_circuit.add_qubit(Qubit("temp_q", process_circuit.n_qubits))

    rename_map_pc = {}
    for index, qb in enumerate(process_circuit.qubits):
        rename_map_pc[qb] = Qubit(index)
    process_circuit.rename_units(cast(Dict[UnitID, UnitID], rename_map_pc))

    pbox = CircBox(process_circuit)

    # set up base circuit for appending xbox to
    base_circuit = Circuit()
    index = 0
    measures = {}
    for qb in all_qubits:
        base_circuit.add_qubit(qb)
        c_bit = Bit(index)
        base_circuit.add_bit(c_bit)
        index += 1
        measures[qb] = c_bit

    # generate state circuits for given correlations
    for major_state_index in range(n_circuits):
        state_circuit = base_circuit.copy()
        # get bit string corresponding to basis state of biggest subset of qubits
        major_state = int_to_binary(major_state_index, major_state_dimensions)
        new_state_dicts = {}
        # parallelise circuits, run uncorrelated subsets characterisation in parallel
        for dim, qubits in zip(subset_dimensions, subsets_matrix_map):
            # add state to prepared states
            new_state_dicts[qubits] = major_state[:dim]
            # find only qubits that are expected to be in 1 state, add xbox to given qubits
            for flipped_qb in itertools.compress(qubits, major_state[:dim]):
                state_circuit.add_circbox(xbox, [cast(UnitID, flipped_qb)])
        # Decompose boxes, add barriers to preserve circuit, add measures
        state_circuit.add_barrier(cast(List[UnitID], all_qubits))

        # add process circuit to measure
        state_circuit.add_circbox(pbox, cast(List[UnitID], state_circuit.qubits))
        DecomposeBoxes().apply(state_circuit)
        state_circuit.add_barrier(cast(List[UnitID], all_qubits))
        for q in measures:
            state_circuit.Measure(q, measures[q])
        # add to returned types
        state_circuit = backend.get_compiled_circuit(state_circuit)
        prepared_circuits.append(state_circuit)
        state_infos.append(StateInfo(new_state_dicts, measures))
    return (prepared_circuits, state_infos)


def calculate_correlation_matrices(
    results_list: List[BackendResult],
    states_info: List[StateInfo],
    correlations: List[List[Node]],
) -> FullCorrelatedNoiseCharacterisation:
    """Calculate the calibration matrices corresponding to some pure noise from the results of running calibration
    circuits.

    :param results_list: List of result via BackendResult. Must be in the same order as the
        corresponding circuits given by prepared_states.
    :param states_info: Each StateInfo object contains the state prepared via a binary
        representation and the qubit_to_bit_map for the corresponding state circuit.
    :param correlations: List of dict corresponding to each prepared basis state

    :return: Characterisation for pure noise given by process circuit
    """

    def to_tuple(correlation_list: List[Node]) -> Tuple[Node, ...]:
        return tuple(correlation_list)

    subsets_matrix_map = OrderedDict.fromkeys(
        sorted(map(to_tuple, correlations), key=len, reverse=True)
    )
    # ordered from largest to smallest via OrderedDict & sorted
    subset_dimensions = [len(subset) for subset in subsets_matrix_map]

    counter = 0
    node_index_dict = dict()
    for qbs, dim in zip(subsets_matrix_map, subset_dimensions):
        # for a subset with n qubits, create a 2^n by 2^n matrix
        subsets_matrix_map[qbs] = np.zeros((1 << dim,) * 2, dtype=float)
        for i in range(len(qbs)):
            qb = qbs[i]
            node_index_dict[qb] = (counter, i)
        counter += 1

    for result, state_info in zip(results_list, states_info):
        state_dict = state_info[0]
        qb_bit_map = state_info[1]
        for qb_sub in subsets_matrix_map:
            # bits of counts to consider
            bits = [qb_bit_map[q] for q in qb_sub]
            counts_dict = result.get_counts(cbits=bits)
            for measured_state, count in counts_dict.items():
                # intended state
                prepared_state_index = binary_to_int(state_dict[qb_sub])
                # produced state
                measured_state_index = binary_to_int(measured_state)
                # update characterisation matrix
                subsets_matrix_map[qb_sub][
                    measured_state_index, prepared_state_index
                ] += count  # type: ignore

    # normalise everything
    normalised_mats = [mat / np.sum(mat, axis=0) for mat in subsets_matrix_map.values()]  # type: ignore
    return FullCorrelatedNoiseCharacterisation(
        correlations, node_index_dict, normalised_mats
    )


# _compute_dot and helper functions #
#
# With thanks to
# https://math.stackexchange.com/a/3423910
# and especially
# https://gist.github.com/ahwillia/f65bc70cb30206d4eadec857b98c4065
# on which this code is based.
def _unfold(tens: np.ndarray, mode: int, dims: List[int]) -> np.ndarray:
    """
    Unfolds tensor into matrix.

    :param tens: Tensor with shape equivalent to dimensions
    :param mode: Specifies axis move to front of matrix in unfolding of tensor
    :param dims: Gives shape of tensor passed

    :return: Matrix with shape (dims[mode], prod(dims[/mode]))
    """
    if mode == 0:
        return tens.reshape(dims[0], -1)
    else:
        return np.moveaxis(tens, mode, 0).reshape(dims[mode], -1)


def _refold(vec: np.ndarray, mode: int, dims: List[int]) -> np.ndarray:
    """
    Refolds vector into tensor.

    :param vec: Tensor with length equivalent to the product of dimensions given in dims
    :param mode: Axis tensor was unfolded along
    :param dims: Shape of tensor

    :return: Tensor folded from vector with shape equivalent to dimensions given in dims
    """
    if mode == 0:
        return vec.reshape(dims)
    else:
        # Reshape and then move dims[mode] back to its
        # appropriate spot (undoing the `unfold` operation).
        tens = vec.reshape([dims[mode]] + [d for m, d in enumerate(dims) if m != mode])
        return np.moveaxis(tens, 0, mode)


def _compute_dot(submatrices: Iterable[np.ndarray], vector: np.ndarray) -> np.ndarray:
    """
    Multiplies the kronecker product of the given submatrices with given vector.

    :param submatrices: Submatrices multiplied
    :param vector: Vector multplied

    :return: Kronecker product of arguments
    """
    dims = [A.shape[0] for A in submatrices]
    vt = vector.reshape(dims)
    for i, A in enumerate(submatrices):
        vt = _refold(A @ _unfold(vt, i, dims), i, dims)
    return vt.ravel()


def _bayesian_iteration(
    submatrices: Iterable[np.ndarray],
    measurements: np.ndarray,
    t: np.ndarray,
    epsilon: float,
) -> np.ndarray:
    """
    Transforms T corresponds to a Bayesian iteration, used to modfiy measurements.

    :param submatrices: submatrices to be inverted and applied to measurements.
    :param measurements: Probability distribution over some set of states to be amended.
    :param t: Some transform to act on measurements.
    :param epsilon: A stabilization parameter  to define an affine transformation for applicatoin
    to submatrices, eliminating zero probabilities.

    :return: Transformed distribution vector.
    """
    # Transform t according to the Bayesian iteration
    # The parameter epsilon is a stabilization parameter which defines an affine
    # transformation to apply to the submatrices to eliminate zero probabilities. This
    # transformation preserves the property that all columns sum to 1
    if epsilon == 0:
        # avoid copying if we don't need to
        As = submatrices
    else:
        As = [
            epsilon / submatrix.shape[0] + (1 - epsilon) * submatrix
            for submatrix in submatrices
        ]
    z = _compute_dot(As, t)
    if np.isclose(z, 0).any():
        raise ZeroDivisionError
    return cast(
        np.ndarray, t * _compute_dot([A.transpose() for A in As], measurements / z)
    )


def _bayesian_iterative_correct(
    submatrices: Iterable[np.ndarray],
    measurements: np.ndarray,
    tol: float = 1e-5,
    max_it: Optional[int] = None,
) -> np.ndarray:
    """
    Finds new states to represent application of inversion of submatrices on measurements.
    Converges when update states within tol range of previously tested states.

    :param submatrices: Matrices comprising the pure noise characterisation.
    :param input_vector: Vector corresponding to some counts distribution.
    :param tol: tolerance of closeness of found results
    :param max_it: Maximum number of inversions attempted to correct results.
    """
    # based on method found in https://arxiv.org/abs/1910.00129

    vector_size = measurements.size
    # uniform initial
    true_states = np.full(vector_size, 1 / vector_size)
    prev_true = true_states.copy()
    converged = False
    count = 0
    epsilon: float = 0  # stabilization parameter, adjusted dynamically
    while not converged:
        if max_it:
            if count >= max_it:
                break
            count += 1
        try:
            true_states = _bayesian_iteration(
                submatrices, measurements, true_states, epsilon
            )
            converged = np.allclose(true_states, prev_true, atol=tol)
            prev_true = true_states.copy()
        except ZeroDivisionError:
            # Shift the stabilization parameter up a bit (always < 0.5).
            epsilon = 0.99 * epsilon + 0.01 * 0.5

    return true_states


class CorrectionMethod(Enum):
    def Invert(
        submatrices: Iterable[np.ndarray], input_vector: np.ndarray
    ) -> np.ndarray:
        """
        Multiplies the kronecker product of given submatrices on input vector
        and then adjusts output to make them genuine probabilities. Submatrices
        represent pure noise characterisation, vector corresponds to counts
        distribution from circuit and device.

        :param submatrices: Matrices comprising the pure noise characterisation.
        :param input_vector: Vector corresponding to some counts distribution.
        """
        try:
            subinverts = [np.linalg.inv(submatrix) for submatrix in submatrices]
        except np.linalg.LinAlgError:
            raise ValueError(
                "Unable to invert calibration matrix: please re-run "
                "calibration experiments or use an alternative correction method."
            )
        # assumes that order of rows in flattened subinverts equals order of bits in input vector
        v = _compute_dot(subinverts, input_vector)
        # The entries of v will always sum to 1, but they may not all be in the range [0,1].
        # In order to make them genuine probabilities (and thus generate meaningful counts),
        # we adjust them by setting all negative values to 0 and scaling the remainder.
        v[v < 0] = 0
        v /= sum(v)
        return v

    def Bayesian(
        submatrices: Iterable[np.ndarray], input_vector: np.ndarray
    ) -> np.ndarray:
        """
        Computes the product of the invert of submatrices on the given input vector via a an iterative
        Bayesian correction method.

        :param submatrices: Matrices comprising the pure noise characterisation.
        :param input_vector: Vector corresponding to some counts distribution.
        :param tol: tolerance of closeness of found results
        :param max_it: Maximum number of inversions attempted to correct results.
        """
        return _bayesian_iterative_correct(
            submatrices, input_vector, tol=1e-5, max_it=500
        )


def reduce_matrix(indices_to_remove: List[int], matrix: np.ndarray) -> np.ndarray:
    """
    Removes indices from indices_to_remove from binary associated to indexing of matrix,
    producing a new transition matrix.
    To do so, it assigns all transition probabilities as the given state in the remaining
    indices binary, with the removed binary in state 0. This is an assumption on the noise made
    because it is likely that unmeasured qubits will be in that state.

    :param indices_to_remove: Binary index of state matrix is mapping to be removed.
    :param matrix: Transition matrix where indices correspond to some binary state, to have some
    dimension removed.

    :return: Transition matrix with removed entries.
    """

    new_n_qubits = int(log2(matrix.shape[0])) - len(indices_to_remove)
    if new_n_qubits == 0:
        return np.array([])
    bin_map = dict()
    mat_dim = 1 << new_n_qubits
    for index in range(mat_dim):
        # get current binary
        bina = list(int_to_binary(index, new_n_qubits))
        # add 0's to fetch old binary to set values from
        for i in sorted(indices_to_remove):
            bina.insert(i, 0)
        # get index of values
        bin_map[index] = binary_to_int(tuple(bina))

    new_mat = np.zeros((mat_dim,) * 2, dtype=float)
    for i in range(len(new_mat)):
        old_row_index = bin_map[i]
        for j in range(len(new_mat)):
            old_col_index = bin_map[j]
            new_mat[i, j] = matrix[old_row_index, old_col_index]
    return new_mat


def reduce_matrices(
    entries_to_remove: List[Tuple[int, int]], matrices: List[np.ndarray]
) -> List[np.ndarray]:
    """
    Removes some dimensions from some matrices.
    :param entries_to_remove: Via indexing, says which dimensions to remove from which indices.
    :param matrices: All matrices to have dimensions removed.

    :return: Matrices with some dimensions removed.
    """
    organise: Dict[int, List[int]] = {k: [] for k in range(len(matrices))}
    for unused in entries_to_remove:
        # unused[0] is index in matrices
        # unused[1] is qubit index in matrix
        organise[unused[0]].append(unused[1])
    output_matrices = [reduce_matrix(organise[m], matrices[m]) for m in organise]
    normalised_mats = [
        mat / np.sum(mat, axis=0) for mat in [x for x in output_matrices if x.size > 0]
    ]
    return normalised_mats


def get_single_matrix(
    entry_to_keep: Tuple[int, int], matrices: List[np.ndarray]
) -> np.ndarray:
    """
    Returns a correction matrix just for the index given.
    :param entry_to_keep: Which matrix and indexing to return a correction matrix for.
    :param matrices: All matrices to find returned matrix from.

    :return: Matrix for correcting given entry.
    """
    mat = matrices[entry_to_keep[0]]
    all_indices = list(range(int(log2(mat.shape[0]))))
    all_indices.remove(entry_to_keep[1])
    return reduce_matrix(all_indices, mat)


def correct_transition_noise(
    result: BackendResult,
    bit_qb_info: Tuple[Dict[Qubit, Bit], Dict[Bit, Qubit]],
    noise_characterisation: FullCorrelatedNoiseCharacterisation,
    corr_method: CorrectionMethod,
) -> BackendResult:
    """
    Modifies count distribution for result, such that the inversion of the pure noise map represented by
    matrices in noise_characterisation is applied to it

    :param result: BackendResult object to be negated by pure noise object.
    :param bit_qb_info: Used to permute corresponding BackendResult object so counts order matches noise characterisation.
    :param noise_characterisation: Object holding all required information for some full noise characterisation of correlated subsets.
    """

    final_measures_qb_map = bit_qb_info[0]
    mid_circuit_measures_bq_map = bit_qb_info[1]
    # get counts from with order of bits that matches order of qubits in subsets
    # if qubit in subset has no bit, skip it
    char_bits_order = []
    unused_final_qbs = []
    for subset in noise_characterisation.CorrelatedNodes:
        for q in subset:
            if q in final_measures_qb_map:
                char_bits_order.append(final_measures_qb_map[q])
            else:
                unused_final_qbs.append(q)

    mid_measure_qbs = []
    for bit in mid_circuit_measures_bq_map:
        mid_measure_qbs.append(mid_circuit_measures_bq_map[bit])
        char_bits_order.append(bit)

    # get counts object for returning later
    counts = result.get_counts(cbits=char_bits_order)
    in_vec = np.zeros(1 << len(char_bits_order), dtype=float)
    # turn from counts to probability distribution
    for state, count in counts.items():
        in_vec[binary_to_int(state)] = count
    Ncounts = np.sum(in_vec)
    in_vec_norm = in_vec / Ncounts
    # remove correlated qubits with no bits in results from matrices for correcting
    # this step just for measures in "final" stage
    reduced_matrices = reduce_matrices(
        [noise_characterisation.NodeToIntDict[q] for q in unused_final_qbs],
        noise_characterisation.CharacterisationMatrices,
    )
    # now added additional matrices to list for correct midcircuit measurements
    for q in mid_measure_qbs:
        reduced_matrices.append(
            get_single_matrix(
                noise_characterisation.NodeToIntDict[q],
                noise_characterisation.CharacterisationMatrices,
            )
        )

    # with counts and characterisation matrices orders matching, correct distribution
    # enum type not callable, ignore
    outvec = corr_method(reduced_matrices, in_vec_norm)  # type: ignore
    outvec *= Ncounts
    # counter object with binary from distribution
    corrected_counts = {
        int_to_binary(index, len(char_bits_order)): Bcount
        for index, Bcount in enumerate(outvec)
    }
    counter = Counter(
        {
            OutcomeArray.from_readouts([key]): ceil(val)
            for key, val in corrected_counts.items()
        }
    )
    # produce and return BackendResult object
    return BackendResult(counts=counter, c_bits=char_bits_order)



"""Visualise TaskGraph using graphviz."""

from typing import TYPE_CHECKING, Iterable, Optional, Tuple

import graphviz as gv  # type: ignore

from qermit.taskgraph.mittask import IOTask

if TYPE_CHECKING:
    import networkx as nx  # type: ignore

    from qermit.taskgraph.mittask import MitTask

# qermit colours
_COLOURS = {
    "background": "white",
    "node": "#c0dd8e",
    "edge": "#6bb24d",
    "dark": "black",
    "node_border": "white",
    "port_border": "#6bb24d",
}
_FONTFACE = "monospace"

_HTML_LABEL_TEMPLATE = """
<TABLE BORDER="{border_width}" CELLBORDER="0" CELLSPACING="1" CELLPADDING="1" BGCOLOR="{node_back_color}" COLOR="{border_colour}">

{inputs_row}

    <TR>
        <TD>
            <TABLE BORDER="0" CELLBORDER="0">
                <TR>
                    <TD><FONT POINT-SIZE="11.0" FACE="{fontface}" COLOR="{label_color}"><B>{node_label}</B></FONT></TD>
                </TR>
            </TABLE>
        </TD>
    </TR>

{outputs_row}

</TABLE>
"""


def _format_html_label(**kwargs):
    _HTML_LABEL_DEFAULTS = {
        "label_color": _COLOURS["dark"],
        "node_back_color": _COLOURS["node"],
        "inputs_row": "",
        "outputs_row": "",
        "border_colour": _COLOURS["port_border"],
        "border_width": "1",
        "fontface": _FONTFACE,
    }
    return _HTML_LABEL_TEMPLATE.format(**{**_HTML_LABEL_DEFAULTS, **kwargs})


_HTML_PORTS_ROW_TEMPLATE = """
    <TR>
        <TD>
            <TABLE BORDER="0" CELLBORDER="0" CELLSPACING="3" CELLPADDING="2">
                <TR>
                    {port_cells}
                </TR>
            </TABLE>
        </TD>
    </TR>
"""

_HTML_PORT_TEMPLATE = (
    '<TD BGCOLOR="{back_colour}" COLOR="{border_colour}"'
    ' PORT="{port}" BORDER="{border_width}">'
    '<FONT POINT-SIZE="10.0" FACE="{fontface}" COLOR="{font_colour}">{port_name}</FONT></TD>'
)


def _html_ports(ports: Iterable[str]) -> str:
    return _HTML_PORTS_ROW_TEMPLATE.format(
        port_cells="".join(
            _HTML_PORT_TEMPLATE.format(
                port_name=port.replace("in_", "").replace("out_", ""),
                port=port,
                back_colour=_COLOURS["background"],
                font_colour=_COLOURS["dark"],
                border_width="1",
                border_colour=_COLOURS["port_border"],
                fontface=_FONTFACE,
            )
            for port in ports
        )
    )


def _node_features(node_name: str, node: "MitTask") -> Tuple[str, str]:
    """Calculate node label (first) and colour (second)."""

    fillcolor = _COLOURS["node"]
    node_label = node_name

    if isinstance(node, IOTask):
        fillcolor = _COLOURS["background"]

    return node_label, fillcolor


def _convert_nodename(node: "MitTask") -> str:
    if node == IOTask.Input:
        return "Inputs"
    if node == IOTask.Output:
        return "Outputs"
    return node._label.replace("::", "-").replace("<", "").replace(">", "")


def _taskgraph_to_graphviz(
    nx_graph: "nx.MultiDiGraph",
    initial_graph: Optional[gv.Digraph] = None,
    label: str = "MitRes",
    prefix: str = "",
) -> gv.Digraph:
    gv_graph = (
        gv.Digraph(
            label,
            strict=False,  # stops multiple shared edges being merged
        )
        if initial_graph is None
        else initial_graph
    )
    graph_atrr = {
        "rankdir": "",
        "ranksep": "0.1",
        "nodesep": "0.15",
        "margin": "0",
        "bgcolor": _COLOURS["background"],
    }
    gv_graph.attr(**graph_atrr)

    for node in nx_graph.nodes:
        node_identifier = prefix + _convert_nodename(node)
        node_name = node_identifier

        node_label, fillcolor = _node_features(node_name, node)

        # node is a table
        # first row is a single cell containing a single row table of inputs
        # second row is table containing single cell of node_label
        # third row is single cell containing a single row table of outputs

        in_ports = [f"in_{i}" for i in range(nx_graph.in_degree(node))]
        out_ports = [f"out_{i}" for i in range(nx_graph.out_degree(node))]

        html_label = _format_html_label(
            node_back_color=fillcolor,
            node_label=node_label,
            inputs_row=_html_ports(in_ports) if in_ports else "",
            outputs_row=_html_ports(out_ports) if out_ports else "",
            border_colour=_COLOURS["background"]
            if fillcolor == _COLOURS["background"]
            else _COLOURS["node_border"],
        )
        gv_graph.node(
            node_identifier,
            label=f"<{html_label}>",
            shape="plain",
        )

    edge_attr = {
        "penwidth": "1.5",
        "arrowhead": "none",
        "arrowsize": "1.0",
        "fontname": _FONTFACE,
        "fontsize": "9",
        "color": _COLOURS["edge"],
        "fontcolor": "black",
    }
    for edge in nx_graph.edges:
        src_node, tgt_node, _ = edge
        src_port = edge[2][0]
        tgt_port = edge[2][1]

        src_nodename = f"{_convert_nodename(src_node)}:out_{src_port}"
        tgt_nodename = f"{_convert_nodename(tgt_node)}:in_{tgt_port}"
        gv_graph.edge(
            src_nodename,
            tgt_nodename,
            **edge_attr,
        )

    return gv_graph



from itertools import combinations
from typing import Any

import networkx as nx  # type: ignore


def _subdag_nodes(subdag: int, node_subdag: dict[Any, int]) -> list[Any]:
    """Get all nodes in a given sub-DAG.

    :param subdag: The sub-DAG whose nodes should be retrieved.
    :param node_subdag: Map from node to the sub-DAG to which the
        node belongs.
    :return: List of nodes belonging to given sub-DAG.
    """
    return [node for node, s in node_subdag.items() if s == subdag]


def _subdag_predecessors(
    dag: nx.DiGraph, subdag: int, node_subdag: dict[Any, int]
) -> list[Any]:
    """Retrieve all nodes not in given sub-DAG with successors in sub-DAG.

    :param dag: Directed Acyclic Graph.
    :param subdag: Sub-DAG to retrieve predecessors of.
    :param node_subdag: Map from node to the sub-DAG to which it belongs.
    :return: Nodes with successors in given sub-DAG.
    """
    subdag_nodes = _subdag_nodes(subdag=subdag, node_subdag=node_subdag)
    return sum(
        [
            [
                predecessor
                for predecessor in dag.predecessors(node)
                # Exclude nodes in subdag.
                if predecessor not in subdag_nodes
            ]
            for node in subdag_nodes
        ],
        start=[],
    )


def _subdag_successors(
    dag: nx.DiGraph, subdag: int, node_subdag: dict[Any, int]
) -> list[Any]:
    """Retrieve all nodes not in given sub-DAG with predecessors in sub-DAG.

    :param dag: Directed Acyclic Graph.
    :param subdag: Sub-DAG to retrieve successors of.
    :param node_subdag: Map from node to the sub-DAG to which it belongs.
    :return: Nodes with predecessors in given sub-DAG.
    """
    subdag_nodes = _subdag_nodes(subdag=subdag, node_subdag=node_subdag)
    return sum(
        [
            [
                successor
                for successor in dag.successors(node)
                # Exclude nodes in subdag.
                if successor not in subdag_nodes
            ]
            for node in subdag_nodes
        ],
        start=[],
    )


def get_monochromatic_convex_subdag(
    dag: nx.DiGraph, coloured_nodes: list[Any]
) -> dict[Any, int]:
    """Retrieve assignment of coloured nodes to sub-DAGs.
    The assignment aims to minimise the number of sub-DAGs.

    :param dag: Directed Acyclic Graph.
    :param coloured_nodes: The nodes which are coloured.
    :return: Map from node to the sub-DAG to which it belongs.
    """

    node_descendants = {node: nx.descendants(dag, node) for node in dag.nodes}
    for node in node_descendants.keys():
        node_descendants[node].add(node)

    def _can_merge(
        dag: nx.DiGraph,
        subdag_one: int,
        subdag_two: int,
        node_subdag: dict[Any, int],
    ) -> bool:
        """Determine if two sub-DAGs can be merged. This will be the case if
        there are no paths between predecessors of one sub-DAG to successors of
        the other.

        :param dag: Directed Acyclic Graph.
        :param subdag_one: First sub-DAG.
        :param subdag_two: Second sub-DAG.
        :param node_subdag: Map from node to the sub-DAG to which it belongs.
        :return: Boolean value indicating if the two sub-DAGs can be merged.
        """

        subdag_two_pred = _subdag_predecessors(
            dag=dag, subdag=subdag_two, node_subdag=node_subdag
        )
        for subdag_one_succ in _subdag_successors(
            dag=dag, subdag=subdag_one, node_subdag=node_subdag
        ):
            if any(
                descendant in subdag_two_pred
                for descendant in node_descendants[subdag_one_succ]
            ):
                return False

        subgraph_one_pred = _subdag_predecessors(
            dag=dag, subdag=subdag_one, node_subdag=node_subdag
        )
        for subdag_two_succ in _subdag_successors(
            dag=dag, subdag=subdag_two, node_subdag=node_subdag
        ):
            if any(
                descendant in subgraph_one_pred
                for descendant in node_descendants[subdag_two_succ]
            ):
                return False

        return True

    node_subdag = {node: i for i, node in enumerate(coloured_nodes)}

    subgraph_merged = True
    while subgraph_merged:
        subgraph_merged = False

        # Try to merge all pairs of sub-DAGs
        for subdag_one, subdag_two in combinations(set(node_subdag.values()), 2):
            if _can_merge(
                dag=dag,
                subdag_one=subdag_one,
                subdag_two=subdag_two,
                node_subdag=node_subdag,
            ):
                for node in _subdag_nodes(subdag=subdag_two, node_subdag=node_subdag):
                    node_subdag[node] = subdag_one
                subgraph_merged = True
                break

    return node_subdag



from typing import cast

from pytket import Circuit, OpType, Qubit
from pytket.passes import BasePass, CustomPass
from pytket.pauli import Pauli

from .noise_model import NoiseModel


def PauliErrorTranspile(noise_model: NoiseModel) -> BasePass:
    """Generates compiler pass which adds coherent noise to a circuit.

    :param noise_model: Model describing the noise to be added. Should be
        a Pauli noise model.
    :return: Compiler pass adding random coherent Pauli noise.
    """

    def add_gates(circuit: Circuit) -> Circuit:
        """Function adding random coherent Pauli errors to a circuit.

        :param circuit: Circuit to which errors are added.
        :raises Exception: Raised if the noise model is not a Pauli one.
        :return: Circuit with additional noise operations.
        """

        # Initialise circuit with the same registers as input.
        noisy_circuit = Circuit()
        for q_register in circuit.q_registers:
            noisy_circuit.add_q_register(q_register)
        for c_register in circuit.c_registers:
            noisy_circuit.add_c_register(c_register)

        # Add each command in the original circuit,
        # and a pauli error if appropriate.
        for command in circuit.get_commands():
            if command.op.type == OpType.Barrier:
                noisy_circuit.add_barrier(command.args)
            else:
                noisy_circuit.add_gate(command.op, command.args)

            # If command has noise model defined, add a random error
            if command.op.type in noise_model.noisy_gates:
                # Sample a random error, which may be None
                error = noise_model.get_error_distribution(command.op.type).sample()
                if error is not None:
                    for qubit, pauli in zip(command.args, error):
                        if pauli in [Pauli.X, OpType.X]:
                            noisy_circuit.X(cast(Qubit, qubit), opgroup="noisy")
                        elif pauli in [Pauli.Z, OpType.Z]:
                            noisy_circuit.Z(cast(Qubit, qubit), opgroup="noisy")
                        elif pauli in [Pauli.Y, OpType.Y]:
                            noisy_circuit.Y(cast(Qubit, qubit), opgroup="noisy")
                        elif pauli in [Pauli.I]:
                            pass
                        else:
                            raise Exception(
                                "Not a Pauli noise model." + f" Contains {pauli} error"
                            )

        return noisy_circuit

    return CustomPass(add_gates)



# Copyright 2019-2023 Quantinuum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import qiskit_aer.noise as noise  # type: ignore
from pytket import OpType
from pytket.extensions.qiskit import AerBackend


class NoisyAerBackend(AerBackend):
    """AerBacked with simple depolarising and SPAM noise model. Depolarising
    noise is added to the gateset {OpType.CX, OpType.H, OpType.Rz,
    OpType.Rz, OpType.Measure} and circuits should be rebased into
    that gateset before running.
    """

    noisy_gate_set = {OpType.CX, OpType.H, OpType.Rz, OpType.Rz, OpType.Measure}

    def __init__(self, n_qubits: int, prob_1: float, prob_2: float, prob_ro: float):
        """AerBacked with simple depolarising and SPAM noise model.

        :param n_qubits: The number of qubits available on the backend.
        :param prob_1: The depolarising noise error rates on single qubit gates.
        :param prob_2: The depolarising noise error rates on two qubit gates.
        :param prob_ro: Error rates of symmetric uncorrelated SPAM errors.
        """

        super().__init__(
            noise_model=self.depolarizing_noise_model(n_qubits, prob_1, prob_2, prob_ro)
        )

    def depolarizing_noise_model(
        self,
        n_qubits: int,
        prob_1: float,
        prob_2: float,
        prob_ro: float,
    ) -> noise.NoiseModel:
        """Generates noise model, may be passed to `noise_model` parameter of
        AerBacked.

        :param n_qubits: Number of qubits noise model applies to.
        :param prob_1: The depolarising noise error rates on single qubit gates.
        :param prob_2: The depolarising noise error rates on two qubit gates.
        :param prob_ro: Error rates of symmetric uncorrelated SPAM errors.
        :return: Noise model
        """

        noise_model = noise.NoiseModel()

        error_2 = noise.depolarizing_error(prob_2, 2)
        for edge in [[i, j] for i in range(n_qubits) for j in range(i)]:
            noise_model.add_quantum_error(error_2, ["cx"], [edge[0], edge[1]])
            noise_model.add_quantum_error(error_2, ["cx"], [edge[1], edge[0]])

        error_1 = noise.depolarizing_error(prob_1, 1)
        for node in range(n_qubits):
            noise_model.add_quantum_error(error_1, ["h", "rx", "rz"], [node])

        probabilities = [[1 - prob_ro, prob_ro], [prob_ro, 1 - prob_ro]]
        error_ro = noise.ReadoutError(probabilities)
        for i in range(n_qubits):
            noise_model.add_readout_error(error_ro, [i])

        return noise_model



# Copyright 2019-2023 Quantinuum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import random
import warnings
from copy import copy
from enum import Enum
from typing import List, Tuple, cast

import numpy as np
from pytket import Circuit, OpType
from pytket.backends import Backend
from pytket.passes import AutoRebase, DecomposeBoxes
from pytket.unit_id import UnitID
from pytket.utils import QubitPauliOperator, get_operator_expectation_value

from qermit import (
    AnsatzCircuit,
    MitEx,
    MitTask,
    ObservableExperiment,
    ObservableTracker,
    SymbolsDict,
    TaskGraph,
)
from qermit.taskgraph import gen_compiled_MitRes

from .cdr_post import (
    _PolyCDRCorrect,
    cdr_calibration_task_gen,
    cdr_correction_task_gen,
    cdr_quality_check_task_gen,
)

ufr_gateset = {OpType.CX, OpType.Rz, OpType.H}
ufr_rebase = AutoRebase(ufr_gateset)


class LikelihoodFunction(Enum):
    def none(
        self, qpo_noisy: QubitPauliOperator, qpo_exact: QubitPauliOperator
    ) -> float:
        """
        Returns probability 1 of accepting returned results.

        :param qpo_noisy: Results calculated from device of choice.
        :param qpo_exact: Results calculated from noiseless simulator of choice.

        :return: Always 1, meaning any result is accepted.
        """
        return 1


def sample_weighted_clifford_angle(rz_angle: float, **kwargs) -> float:
    """
    Calculates a weights distribution over different possible Clifford gates from input gate.
    Clifford gates prepared by taking S gate to the power of n in {0,4}.
    n value sampled from calculated weights distribution.
    Distribution calculation as in B1, page 6 arXiv:2005.10189.

    :param rz_angle: Angle of rotation in rz axis.
    :key: seed

    :return: An angle corresponding to Clifford rotation of some Rz gate
    """
    if "seed" in kwargs:
        random.seed(kwargs.get("seed"))

    rz_angle = rz_angle % 2
    rz_angle_matrix = np.asarray(
        [
            [np.exp(-0.5 * np.pi * rz_angle * 1j), 0],
            [0, np.exp(0.5 * np.pi * rz_angle * 1j)],
        ]
    )
    weights = []
    for n in range(4):
        sn_matrix = np.asarray(
            [[np.exp(-0.25 * np.pi * n * 1j), 0], [0, np.exp(0.25 * np.pi * n * 1j)]]
        )
        d = np.linalg.norm(rz_angle_matrix - sn_matrix)
        weights.append(np.exp((-(d**2)) * 4))
    return 0.5 * random.choices(range(4), weights)[0]


def gen_state_circuits(
    c: Circuit, n_non_cliffords: int, n_pairs: int, total_state_circuits: int, **kwargs
) -> List[Circuit]:
    """
    For given circuit c, returns total_state_circuits number of circuits, where each circuit is
    run through some MitEx object to provide characterisation data for later correction.

    State circuit construction as in appendix B of arXiv:2005.10189.
    State circuits are generated via a Markov Chain Monte Carlo technique.
    Circuit c is first rebased into a basis set of CX, H and Rz gates.
    Circuit c is then modified to a near Clifford Circuit with only n_non_cliffords
    number of Rz gates with non-Clifford angles.
    The near Clifford circuit is generated by randomly choosing n_non_cliffords non-Clifford Rz
    gates in c to retain their angle, and then replacing all other non-Clifford
    Rz gates with some random Clifford gate.
    Then, for each update step to generate a new state circuit the following occurs:
    • n_pairs of pairs of gates in c are randomly chosen
    • each pair consists of one Rz gate with non-Clifford angle (amongst the n_non_cliffords)
    and one Rz gate with Clifford angle that was originally non-Clifford.
    • The Rz gate with Clifford angle has its angle replaced with its original non-Clifford angle
    • The non-Clifford Rz gate angle has its angle replaced with a Clifford angle
    • A Metropolis-Hastings rule is used to accept or reject this new state circuit late when results gathered


    :param c: Circuit for producing state circuits from.
    :param n_non_cliffords: Number of non-Clifford gates in resulting characterisation state circuits
    :param n_pairs: Pairs of Clifford, Non-Clifford gates in state circuit generated.
    :param total_state_circuits: Total number of state circuits to be produced for characterisation
    :key: seed for random methods

    :return: All generated state circuits
    """
    # set seed if given
    if "seed" in kwargs:
        random.seed(kwargs.get("seed"))
        np.random.seed(kwargs.get("seed"))

    # Work in CX, H, Rz basis for ease
    DecomposeBoxes().apply(c)
    ufr_rebase.apply(c)
    c.flatten_registers()
    all_coms = c.get_commands()

    #  angles that make Clifford gates for S^n
    clifford_angles = set({0, 0.5, 1.0, 1.5, 2, 2.5, 3, 3.5})

    if n_pairs > n_non_cliffords:
        raise ValueError(
            "More pairs {} than total non-clifford gates {}. Number of pairs must be less than or equal to.".format(
                n_pairs, n_non_cliffords
            )
        )
    # Admin for circuit modifications
    # Create a register of ints corresponding to indices of commands list with non-Clifford Rz gates
    rz_ops = set()
    for i in range(len(all_coms)):
        if all_coms[i].op.type == OpType.Rz:
            if all_coms[i].op.params[0] not in clifford_angles:
                rz_ops.add(i)
    if len(rz_ops) == 0:
        return [c] * total_state_circuits

    state_circuits: List[Circuit] = []
    if len(rz_ops) == 1:
        # make special case where its not possible to swap pairs
        # produce state circuits until limit reached
        while len(state_circuits) < total_state_circuits:
            new_circuit = Circuit(c.n_qubits, len(c.bits))
            for i in range(len(all_coms)):
                com = all_coms[i]
                if com.op.type == OpType.Rz:
                    new_circuit.add_barrier(cast(List[UnitID], com.qubits))
                    original_angle = com.op.params[0]
                    if not isinstance(original_angle, float):
                        raise Exception(
                            "Circuit cannot include parameters which are not floats."
                        )
                    angle = sample_weighted_clifford_angle(original_angle)
                    new_circuit.add_gate(
                        com.op.type, [angle], cast(List[UnitID], com.qubits)
                    )
                    new_circuit.add_barrier(cast(List[UnitID], com.qubits))
                # Measure gate has special case, but can assume 1 qubit to 1 bit
                elif com.op.type is OpType.Measure:
                    new_circuit.Measure(com.qubits[0], com.bits[0])
                # A special case for Barrier metaop
                elif com.op.type == OpType.Barrier:
                    new_circuit.add_barrier(com.args)
                # CX or H gate, add as is
                else:
                    new_circuit.add_gate(com.op.type, cast(List[UnitID], com.qubits))

            # all circuits accepted and run, some results later discarded if not accepted by Metropolis-Hastings rule
            state_circuits.append(new_circuit)
        return state_circuits

    # reassign variables where appropriate to guarantee there is always one pair of non-Clifford Clifford
    # that can be resampled
    n_non_cliffords = min(n_non_cliffords, len(rz_ops) - 1)
    n_cliffords = len(rz_ops) - n_non_cliffords
    n_pairs = min(n_cliffords, n_non_cliffords, n_pairs)

    # non_cliffords are indices for gates to be left non Clifford
    non_cliffords = np.random.choice(list(rz_ops), n_non_cliffords, replace=False)

    # rz_ops then only contains rz gates in c to be substituted for Clifford angles
    rz_ops.difference_update(non_cliffords)
    # Power of random Clifford gates to be substituted
    cliffords = {num: random.randint(0, 4) for num in rz_ops}

    # keep on producing state circuits until limit reached
    while len(state_circuits) < total_state_circuits:
        # cliffords.keys() are integers for now Clifford gates
        # sample some set of these to be subbed for original non-Clifford angle
        clifford_pair_elements = random.sample(list(cliffords.keys()), n_pairs)
        # from remaining non-Clifford Rz gates, sample some to have random Clifford gate
        non_clifford_pair_elements = random.sample(list(non_cliffords), n_pairs)

        # create new Circuit from scratch
        new_circuit = Circuit(c.n_qubits, len(c.bits))
        for i in range(len(all_coms)):
            com = all_coms[i]
            if com.op.type == OpType.Rz:
                new_circuit.add_barrier(cast(List[UnitID], com.qubits))
                # 3 sets of gates int must be in
                # in clifford_pair_elements means gate has been denominated as Clifford,
                # but is in some sampled pair so add original angle
                if i in clifford_pair_elements:
                    new_circuit.add_gate(
                        com.op.type, com.op.params, cast(List[UnitID], com.qubits)
                    )
                    new_circuit.add_barrier(cast(List[UnitID], com.qubits))
                # in non_clifford_pair_elements mean gate was denominated to be left non-Clifford,
                # but its value has been sampled in a pair to now be Clifford
                # random angle is sampled and returned
                elif i in non_clifford_pair_elements:
                    original_angle = com.op.params[0]
                    if not isinstance(original_angle, float):
                        raise Exception(
                            "Circuit cannot include parameters which are not floats."
                        )
                    angle = sample_weighted_clifford_angle(original_angle)
                    new_circuit.add_gate(
                        com.op.type, [angle], cast(List[UnitID], com.qubits)
                    )
                    new_circuit.add_barrier(cast(List[UnitID], com.qubits))
                # in cliffords mean it is denominated as Clifford, and hasn't been sampled for a pair
                # as clifford_pair_elements has already been checked
                # in this case, cliffords is a dict between Rz index and substitution S power
                # get power from dict, multiply by 0.5 to get angle, add to circuit
                elif i in cliffords:
                    new_circuit.add_gate(
                        com.op.type,
                        [0.5 * cliffords[i]],
                        cast(List[UnitID], com.qubits),
                    )
                    new_circuit.add_barrier(cast(List[UnitID], com.qubits))
                # final case means gate was chosen to retain non-Clifford, and has not been
                # sampled in any pair, so add original angle.
                else:
                    new_circuit.add_gate(
                        com.op.type, com.op.params, cast(List[UnitID], com.qubits)
                    )
                    new_circuit.add_barrier(cast(List[UnitID], com.qubits))
            # Measure gate has special case, but can assume 1 qubit to 1 bit
            elif com.op.type is OpType.Measure:
                new_circuit.Measure(com.qubits[0], com.bits[0])
            # A special case for Barrier metaop
            elif com.op.type == OpType.Barrier:
                new_circuit.add_barrier(com.args)
            # CX or H gate, add as is
            else:
                new_circuit.add_gate(com.op.type, cast(List[UnitID], com.qubits))

        # all circuits accepted and run, some results later discarded if not accepted by Metropolis-Hastings rule
        state_circuits.append(new_circuit)

    return state_circuits


def ccl_state_task_gen(
    n_non_cliffords: int,
    n_pairs: int,
    total_state_circuits: int,
    simulator_backend: Backend,
    tolerance: float,
    max_state_circuits_attempts: int,
) -> MitTask:
    """
    Returns a MitTask object for which given some set of experiments,
    for each experiment prepares a set of state circuits for Clifford Circuit Learning characterisation.
    The original experiment is returned on the first wire, state circuits for running on backend on second wire,
    and state circuits for noiseless simulation on the third wire.

    :param n_non_cliffords: Number of remaining non-Clifford gates in generated State Circuits.
    :param n_pairs: Parameter used for guiding properties of State Circuits generated.
    :param total_state_circuits: Number of state circuits prepared for characterisation.
    :param tolerance: Model can be perturbed when calibration circuits have by
        exact expectation values too close to each other. This parameter
        sets a distance between exact expectation values which at least some
        calibration circuits should have.
    :param simulator_backend: Backend object simulated characterisation experiments are
        default run through.
    :param max_state_circuits_attempts: The maximum number of times to attempt to generate a
        list of calibrations circuit with significantly different expectation
        values, before resorting to a list with similar expectation values.

    :return: MitTask object for preparing and returning state circuits for characterisation.
    """

    def task(
        obj,
        experiment_wires: List[ObservableExperiment],
    ) -> Tuple[
        List[ObservableExperiment],
        List[ObservableExperiment],
        List[ObservableExperiment],
    ]:
        """
        :param experiment_wires: Information used to define generic experiments in MitEx objects.

        :return: Original experiment for running on experiment backend, state circuits for running on characterisation backend,
        state circuits for running on noiseless backend.
        """
        simulator_wires = []
        device_wires = []
        for measurement_wire in experiment_wires:
            ansatz_circuit = measurement_wire.AnsatzCircuit
            shots = ansatz_circuit.Shots
            qubit_pauli_operator = (
                measurement_wire.ObservableTracker.qubit_pauli_operator
            )
            # generate all state circuits
            c_copy = ansatz_circuit.Circuit.copy()
            c_copy.symbol_substitution(ansatz_circuit.SymbolsDict._symbolic_map)

            all_close = True
            attempt = 0
            while all_close and attempt < max_state_circuits_attempts:
                state_circuits = gen_state_circuits(
                    c_copy,
                    n_non_cliffords,
                    n_pairs,
                    total_state_circuits,
                )

                pauli_expectation_list = [
                    get_operator_expectation_value(
                        c, qubit_pauli_operator, simulator_backend
                    )
                    for c in state_circuits
                ]
                all_close = all(
                    abs(pauli_expectation - pauli_expectation_list[0]) <= tolerance
                    for pauli_expectation in pauli_expectation_list
                )

                attempt += 1

            if all_close:
                warnings.warn(
                    "Clifford Data Regression performs best when the exact expectation values of all calibration circuits are not the same. However, the generated calibration circuits have similar exact expectation values. Fit of the extrapolation function may be poor as a result."
                )

            # for each state circuit, create a new wire of for each state circuit
            # one for simulator, one for device
            for c in state_circuits:
                wire_sim = ObservableExperiment(
                    AnsatzCircuit=AnsatzCircuit(
                        Circuit=c, Shots=shots, SymbolsDict=SymbolsDict()
                    ),
                    ObservableTracker=ObservableTracker(
                        copy(qubit_pauli_operator)
                    ),  # no copy means changes to one QubitPauliOperator can be made to all
                )
                wire_device = ObservableExperiment(
                    AnsatzCircuit=AnsatzCircuit(
                        Circuit=c.copy(),
                        Shots=copy(shots),
                        SymbolsDict=SymbolsDict(),
                    ),
                    ObservableTracker=ObservableTracker(copy(qubit_pauli_operator)),
                )
                simulator_wires.append(wire_sim)
                device_wires.append(wire_device)
        return (experiment_wires, simulator_wires, device_wires)

    return MitTask(
        _label="CCLStateCircuits",
        _n_in_wires=1,
        _n_out_wires=3,
        _method=task,
    )


def ccl_result_batching_task_gen(n_state_circuits: int) -> MitTask:
    """
    For each experiment run through MitEx, pairs up noisy and noiseless expectation values
    from state circuits for that experiments CCL calibration and then returns results for a single
    calibration in a single list.

    :param n_state_circuits: Number of state circuits initially prepared for each
        experiment characterisation.

    :return: MitTask object that organises QubitPauliOperator objects required for
        characterisation.
    """

    def task(
        obj, exact_exp: List[QubitPauliOperator], noisy_exp: List[QubitPauliOperator]
    ) -> Tuple[List[List[Tuple[QubitPauliOperator, QubitPauliOperator]]]]:
        """
        :param noisy_exp: All QubitPauliOperators returned from running state circuit calibrations for all experiments through device.
        :param exact_exp: All QubitPauliOperators returned from running state circuit calibrations for all experiments through noiseless simulator.

        :return: State circuit results split into separate lists for each experiment, with noisy and noiseless expectations paired together.
        """
        if len(noisy_exp) != len(exact_exp):
            raise RuntimeError(
                "Batching task should receive identical number of Simulated and Device run results."
            )

        zipped = list(zip(noisy_exp, exact_exp))
        chunked_zipped = [
            zipped[i : i + n_state_circuits]
            for i in range(0, len(zipped), n_state_circuits)
        ]
        return (chunked_zipped,)

    return MitTask(
        _label="CCLBatchResults", _n_in_wires=2, _n_out_wires=1, _method=task
    )


def ccl_likelihood_filtering_task_gen(
    likelihood_function: LikelihoodFunction, **kwargs
) -> MitTask:
    """
    :param likelihood_function: LikelihoodFunction enum used to accept or reject some pair of noisy and noiseless expectation.
        Function must take two QubitPauliOperator as parameter, and return a single float between 0 and 1 as answer.
    :key seed: Seed value for sampling probability for likelihood function

    :return: MitTask object that removes some characterisation results under some
        condition set by the likelihood_function option.
    """

    def task(
        obj,
        state_circuit_exp: List[List[Tuple[QubitPauliOperator, QubitPauliOperator]]],
    ) -> Tuple[List[List[Tuple[QubitPauliOperator, QubitPauliOperator]]]]:
        """
        For each combination of noisy and noiseless expectation value
        for some state circuit, use a Metropolis-Hastings rule with
        given likelihood function to accept or reject the result.
        In this manner, this task filters unwanted results, returning only
        accepted expectations for calibrating from.

        :param state_circuit_exp: Noisy and Noiseless Expectation results for calibration.

        :return: Filtered calibration results.
        """
        if likelihood_function == LikelihoodFunction.none:
            return (state_circuit_exp,)
        else:
            if "seed" in kwargs:
                random.seed(kwargs.get("seed"))
            filtered_results = []
            for exp in state_circuit_exp:
                filtered_experiment = []
                for noisy, exact in exp:
                    likelihood_res = likelihood_function(noisy, exact)  # type: ignore
                    if random.uniform(0, 1) < likelihood_res:
                        filtered_experiment.append((noisy, exact))
                filtered_results.append(filtered_experiment)
            return (filtered_results,)

    return MitTask(
        _label="CCLLikelihoodFilterResults", _n_in_wires=1, _n_out_wires=1, _method=task
    )


def gen_CDR_MitEx(
    device_backend: Backend,
    simulator_backend: Backend,
    n_non_cliffords: int,
    n_pairs: int,
    total_state_circuits: int,
    **kwargs,
) -> MitEx:
    """
    Produces a MitEx object for applying Clifford Circuit Learning & Clifford Data Regression
    mitigation methods when calculating expectation values of observables. Implementation as
    in arXiv:2005.10189.

    :param device_backend: Backend object device experiments are default run through.
    :param simulator_backend: Backend object simulated characterisation experiments are
        default run through.
    :param n_non_cliffords: Number of gates in Ansatz Circuit left as non-Clifford gates when
        producing characterisation circuits.
    :param n_pairs: Number of non-Clifford gates sampled to become Clifford and vice versa
        each time a new state circuit is generated.
    :param total_state_circuits: Total number of state circuits produced for characterisation.
    :key states_simulator_mitex: MitEx object noiseless characterisation simulations are executed on, default
        simulator_backend with basic compilation of circuit.
    :key states_device_mitex: MitEx object noisy characterisation circuit are executed on, default
        device_backend with basic compilation of circuit.
    :key experiment_mitex: MitEx object that actual experiment circuits are executed on, default
        backend with some compilation of circuit.
    :key model: Model characterised by state circuits, default _PolyCDRCorrect(1) (see cdr_post.py for other options).
    :key likelihood_function: LikelihoodFunction used to filter state circuit results, given by a LikelihoodFunction Enum,
        default set to none.
    :key tolerance: Model can be perturbed when calibration circuits have by
        exact expectation values too close to each other. This parameter
        sets a distance between exact expectation values which at least some
        calibration circuits should have.
    :key distance_tolerance: The absolute tolerance on the distance between
        expectation values of the calibration and original circuit.
    :key calibration_fraction: The upper bound on the fraction of calibration
        circuits which have noisy expectation values far from that of the
        original circuit.
    """
    _states_sim_mitex = copy(
        kwargs.get(
            "states_simluator_mitex",
            MitEx(
                simulator_backend,
                _label="StatesSimMitEx",
                mitres=gen_compiled_MitRes(simulator_backend, 0),
            ),
        )
    )
    _states_device_mitex = copy(
        kwargs.get(
            "states_device_mitex",
            MitEx(
                device_backend,
                _label="StatesDeviceMitEx",
                mitres=gen_compiled_MitRes(device_backend, 0),
            ),
        )
    )
    _experiment_mitex = copy(
        kwargs.get(
            "experiment_mitex",
            MitEx(
                device_backend,
                _label="ExperimentMitEx",
                mitres=gen_compiled_MitRes(device_backend, 0),
            ),
        )
    )

    _states_sim_taskgraph = TaskGraph().from_TaskGraph(_states_sim_mitex)
    _states_sim_taskgraph.parallel(_states_device_mitex)
    _states_sim_taskgraph.append(ccl_result_batching_task_gen(total_state_circuits))

    likelihood_function = kwargs.get("likelihood_function", LikelihoodFunction.none)

    _experiment_taskgraph = TaskGraph().from_TaskGraph(_experiment_mitex)
    _experiment_taskgraph.parallel(_states_sim_taskgraph)

    _post_calibrate_task_graph = TaskGraph(_label="FitCalibrate")
    _post_calibrate_task_graph.append(
        ccl_likelihood_filtering_task_gen(likelihood_function)
    )
    _post_calibrate_task_graph.append(
        cdr_calibration_task_gen(
            kwargs.get("model", _PolyCDRCorrect(1)),
        )
    )

    _post_task_graph = TaskGraph(_label="QualityCheckCorrect")
    _post_task_graph.parallel(_post_calibrate_task_graph)
    _post_task_graph.prepend(
        cdr_quality_check_task_gen(
            distance_tolerance=kwargs.get("distance_tolerance", 0.1),
            calibration_fraction=kwargs.get("calibration_fraction", 0.5),
        )
    )

    _experiment_taskgraph.prepend(
        ccl_state_task_gen(
            n_non_cliffords,
            n_pairs,
            total_state_circuits,
            simulator_backend=simulator_backend,
            tolerance=kwargs.get("tolerance", 0.01),
            max_state_circuits_attempts=kwargs.get("max_state_circuits_attempts", 10),
        )
    )
    _experiment_taskgraph.append(_post_task_graph)
    _experiment_taskgraph.append(cdr_correction_task_gen())

    return MitEx(device_backend).from_TaskGraph(_experiment_taskgraph)



# Copyright 2019-2023 Quantinuum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional, Sequence, Union

from pytket import Circuit, OpType
from pytket.architecture import FullyConnected
from pytket.backends.backendinfo import BackendInfo
from pytket.backends.backendresult import BackendResult
from pytket.backends.resulthandle import ResultHandle
from pytket.extensions.quantinuum import QuantinuumBackend
from pytket.extensions.quantinuum.backends.quantinuum import _ALL_GATES
from pytket.passes import AutoRebase
from pytket.predicates import CompilationUnit, GateSetPredicate

from .noisy_aer_backend import NoisyAerBackend


class MockQuantinuumBackend(QuantinuumBackend):
    """Backend mocking some of the features of QuantinuumBackend.
    In particular the gateset and connectivity of the backend is replicated
    so that compilation behaviour is reproduced. Some noise (unrelated to
    that on the device) is also applied.
    """

    gate_set = _ALL_GATES
    gate_set.add(OpType.ZZPhase)

    backend_info = BackendInfo(
        name="MockQuantinuumBackend",
        device_name="mock-quantinuum",
        version="n/a",
        architecture=FullyConnected(10, "q"),
        gate_set=gate_set,
        n_cl_reg=100,
    )

    def __init__(self):
        super(MockQuantinuumBackend, self).__init__(device_name="H1-1SC")
        self.noisy_backend = NoisyAerBackend(
            self.backend_info.n_nodes, 0.0001, 0.001, 0.01
        )
        self.handle_cu_dict = dict()

    def process_circuit(
        self,
        circuit: Circuit,
        n_shots: Optional[int] = None,
        valid_check: bool = True,
        **kwargs,
    ) -> ResultHandle:
        """Submit circuit to the backend for running.

        :param circuit: Circuit to process on the backend
        :param n_shots: Number of shots to run per circuit.
        :param valid_check: Explicitly check that all circuits satisfy all
            required predicates to run on the backend, defaults to True
        :return: Handles to results for each input circuit, as an interable
            in the same order as the circuits.
        """

        if valid_check:
            assert self.valid_circuit(circuit)

        noisy_circuit = circuit.copy()
        cu = CompilationUnit(noisy_circuit)

        self.noisy_backend.default_compilation_pass(optimisation_level=0).apply(cu)
        AutoRebase(gateset=self.noisy_backend.noisy_gate_set).apply(cu)
        assert GateSetPredicate(
            self.noisy_backend.noisy_gate_set.union({OpType.Reset, OpType.Barrier})
        ).verify(cu.circuit)

        handle = self.noisy_backend.process_circuit(cu.circuit, n_shots)
        self.handle_cu_dict[handle] = cu

        return handle

    def process_circuits(
        self,
        circuits: Sequence[Circuit],
        n_shots: Union[Optional[int], Sequence[Optional[int]]] = None,
        valid_check: bool = True,
        **kwargs,
    ) -> List[ResultHandle]:
        """Submit list of circuits to the backend for running.

        :param circuits: Circuits to process on the backend
        :param n_shot: Number of shots to run per circuit. May be a list of
            the same length as circuits.
        :param valid_check: Explicitly check that all circuits satisfy all
            required predicates to run on the backend, defaults to True
        :return: Handles to results for each input circuit, as an interable in the same order as the circuits.
        """
        if (isinstance(n_shots, int)) or (n_shots is None):
            return [
                self.process_circuit(circuit, n_shots, valid_check=valid_check)
                for circuit in circuits
            ]
        return [
            self.process_circuit(circuit, n_shot, valid_check=valid_check)
            for circuit, n_shot in zip(circuits, n_shots)
        ]

    def get_result(self, handle: ResultHandle, **kwargs) -> BackendResult:
        """Return a BackendResult corresponding to the handle.

        :param handle: handle to results
        :return: Results corresponding to handle.
        """
        return self.noisy_backend.get_result(handle, *kwargs)



from collections import Counter
from typing import Dict, List, Tuple

from pytket.backends.backendresult import BackendResult
from pytket.circuit import Bit
from pytket.utils.outcomearray import OutcomeArray


class PostselectMgr:
    """Class for tracking and applying post selection to results.

    An example use case might be the following. Here a Bell state is
    prepared. We would like to keep one bit of the results, conditioned
    on the other being 0. That's to say that the postselected
    bits should all be 0.

    .. jupyter-execute::

        from pytket import Circuit, Bit, Qubit
        from pytket.circuit.display import render_circuit_jupyter

        # Two qubits. The result of measuring the first will
        # be used to postselect the result of measuring the second.
        post_q = Qubit(0)
        comp_q = Qubit(1)

        # Construct Bell state preparation circuit.
        circuit = Circuit()
        circuit.add_qubit(post_q)
        circuit.add_qubit(comp_q)

        circuit.H(post_q)
        circuit.CX(post_q, comp_q)

        post_b = Bit(0)
        comp_b = Bit(1)

        circuit.add_bit(post_b)
        circuit.add_bit(comp_b)

        circuit.Measure(comp_q, comp_b)
        circuit.Measure(post_q, post_b)

        render_circuit_jupyter(circuit)

    Running this circuit gives a roughly equal mix of 00 and 11
    computation basis states.

    .. jupyter-execute::

        from qermit.postselection import PostselectMgr
        from pytket.extensions.quantinuum import QuantinuumBackend, QuantinuumAPIOffline

        backend = QuantinuumBackend(
            device_name="H1-1LE",
            api_handler = QuantinuumAPIOffline(),
        )

        compiled_circuit = backend.get_compiled_circuit(circuit)
        result = backend.run_circuit(compiled_circuit, 100)
        result.get_counts()

    """

    def __init__(
        self,
        compute_cbits: List[Bit],
        postselect_cbits: List[Bit],
    ):
        """
        This class is straightforwardly initialised with the computation
        and post selection bits. The computation bits are post selected based
        on the results of the post selection bits.

        .. jupyter-execute::

            postselect_mgr = PostselectMgr(
                compute_cbits=[comp_b],
                postselect_cbits=[post_b]
            )

        :param compute_cbits: Bits in the circuit which are not affected
            by post selection.
        :param postselect_cbits: Bits on which the post selection is based.
        :raises Exception: Raised if a bit is in both compute_cbits
            and postselect_cbits.
        """

        intersect = set(compute_cbits).intersection(set(postselect_cbits))
        if intersect:
            raise Exception(
                f"{intersect} are post select and compute qubits. "
                + "They cannot be both."
            )

        self.compute_cbits: List[Bit] = compute_cbits
        self.postselect_cbits: List[Bit] = postselect_cbits

        self.cbits: List[Bit] = compute_cbits + postselect_cbits

    def _get_postselected_shot(self, shot: Tuple[int, ...]) -> Tuple[int, ...]:
        "Removes postselection bits from shot."
        return tuple(
            [
                bit
                for bit, reg in zip(shot, self.cbits)
                if reg not in self.postselect_cbits
            ]
        )

    def _is_postselect_shot(self, shot: Tuple[int, ...]) -> bool:
        "Determines if shot survives postselection"

        # TODO: It may be nice to generalise this so that other functions
        # besides bit==0 can be used as a means of postselection.
        return all(
            bit == 0
            for bit, reg in zip(shot, self.cbits)
            if reg in self.postselect_cbits
        )

    def _dict_to_result(self, result_dict: Dict[Tuple[int, ...], int]) -> BackendResult:
        """Convert dictionary to BackendResult.

        :param result_dict: Dictionary to convert. Should be in the form of
            map from shot to count.
        :return: Corresponding BackendResult.
        """

        # Special case where the dictionary is empty. Presently having
        # an empty counter results in an error.
        if not result_dict:
            return BackendResult()

        return BackendResult(
            counts=Counter(
                {
                    OutcomeArray.from_readouts([key]): val
                    for key, val in result_dict.items()
                }
            ),
            c_bits=self.compute_cbits,
        )

    def postselect_result(self, result: BackendResult) -> BackendResult:
        """Transforms BackendResult to keep only shots which should be
        post selected.

        .. jupyter-execute::

            post_result = postselect_mgr.postselect_result(result=result)
            post_result.get_counts()

        :param result: Result to be modified.
        :return: Postselected shots.
        """

        return self._dict_to_result(
            {
                self._get_postselected_shot(shot): count
                for shot, count in result.get_counts(cbits=self.cbits).items()
                if self._is_postselect_shot(shot)
            }
        )

    def merge_result(self, result: BackendResult) -> BackendResult:
        """Transforms BackendResult so that postselection bits are
        removed, but no shots are removed by postselection.

        .. jupyter-execute::

            merge_result = postselect_mgr.merge_result(result=result)
            merge_result.get_counts()

        :param result: Result to be transformed.
        :return: Result with postselection bits removed.
        """

        merge_dict: Dict[Tuple[int, ...], int] = {}
        for shot, count in result.get_counts(cbits=self.cbits).items():
            postselected_shot = self._get_postselected_shot(shot)
            merge_dict[postselected_shot] = merge_dict.get(postselected_shot, 0) + count

        return self._dict_to_result(merge_dict)



# Copyright 2019-2023 Quantinuum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
The spam module defines tasks for completing SPAM Mitigation methods, and generator
functions for producing MitRes objects that automatically characterise and apply SPAM
Mitigation methods to experiment results.
"""

from .full_transition_tomography import FullCorrelatedNoiseCharacterisation
from .spam_mitres import (
    CorrectionMethod,
    gen_FullyCorrelated_SPAM_MitRes,
    gen_UnCorrelated_SPAM_MitRes,
)
